@@ -25,6 +25,7 @@
 #include <gsf/gsf-impl-utils.h>
 #include <gsf/gsf-utils.h>
 #include <gsf/gsf-msole-impl.h>
+#include <gsf/gsf-input-proxy.h>
 
 #include <string.h>
 #include <stdio.h>
@@ -32,6 +33,8 @@
 #undef G_LOG_DOMAIN
 #define G_LOG_DOMAIN "libgsf:msole"
 
+static GObjectClass *parent_class;
+
 typedef struct {
 	guint32 *block;
 	guint32  num_blocks;
@@ -232,6 +235,8 @@ ole_info_get_sb_file (GsfInfileMSOle *parent)
 
 	parent->info->sb_file = gsf_infile_msole_new_child (parent,
 		parent->info->root_dir, NULL);
+	if (!parent->info->sb_file)
+		return NULL;
 
 	/* avoid creating a circular reference */
 	ole_info_unref (((GsfInfileMSOle *)parent->info->sb_file)->info);
@@ -300,9 +305,9 @@ ole_dirent_new (GsfInfileMSOle *ole, guint32 entry, MSOleDirent *parent)
 		return NULL;
 	}
 
-	/* It looks like directory sizes are sometimes bogus */
+	/* It looks like directory (and root directory) sizes are sometimes bogus */
 	size = GSF_LE_GET_GUINT32 (data + DIRENT_FILE_SIZE);
-	g_return_val_if_fail (type == DIRENT_TYPE_DIR ||
+	g_return_val_if_fail (type == DIRENT_TYPE_DIR || type == DIRENT_TYPE_ROOTDIR ||
 			      size <= (guint32)ole->input->size, NULL);
 
 	dirent = g_new0 (MSOleDirent, 1);
@@ -435,8 +440,12 @@ ole_dup (GsfInfileMSOle const *src, GError **err)
 	g_return_val_if_fail (src != NULL, NULL);
 
 	input = gsf_input_dup (src->input, err);
-	if (input == NULL)
+	if (input == NULL) {
+		if (err != NULL)
+			*err = g_error_new (gsf_input_error (), 0,
+					    "Failed to duplicate input stream");
 		return NULL;
+	}
 
 	dst = (GsfInfileMSOle *)g_object_new (GSF_INFILE_MSOLE_TYPE, NULL);
 	dst->input = input;
@@ -469,7 +478,7 @@ ole_init_info (GsfInfileMSOle *ole, GError **err)
 	guint32 metabat_block, *ptr;
 
 	/* check the header */
-	if (gsf_input_seek (ole->input, (gsf_off_t) 0, G_SEEK_SET) ||
+	if (gsf_input_seek (ole->input, 0, G_SEEK_SET) ||
 	    NULL == (header = gsf_input_read (ole->input, OLE_HEADER_SIZE, NULL)) ||
 	    0 != memcmp (header, signature, sizeof (signature))) {
 		if (err != NULL)
@@ -595,7 +604,6 @@ ole_init_info (GsfInfileMSOle *ole, GError **err)
 static void
 gsf_infile_msole_finalize (GObject *obj)
 {
-	GObjectClass *parent_class;
 	GsfInfileMSOle *ole = GSF_INFILE_MSOLE (obj);
 
 	if (ole->input != NULL) {
@@ -611,9 +619,7 @@ gsf_infile_msole_finalize (GObject *obj)
 
 	g_free (ole->stream.buf);
 
-	parent_class = (GObjectClass *)g_type_class_peek (GSF_INFILE_TYPE);
-	if (parent_class && parent_class->finalize)
-		parent_class->finalize (obj);
+	parent_class->finalize (obj);
 }
 
 static GsfInput *
@@ -728,6 +734,9 @@ gsf_infile_msole_new_child (GsfInfileMSOle *parent,
 	size_t size_guess;
 
 	child = ole_dup (parent, err);
+	if (!child)
+		return NULL;
+
 	child->dirent = dirent;
 	gsf_input_set_size (GSF_INPUT (child), (gsf_off_t) dirent->size);
 
@@ -751,6 +760,13 @@ gsf_infile_msole_new_child (GsfInfileMSOle *parent,
 		metabat = &info->sb.bat;
 		size_guess = dirent->size >> info->sb.shift;
 		sb_file = ole_info_get_sb_file (parent);
+		if (!sb_file) {
+			if (err != NULL)
+				*err = g_error_new (gsf_input_error (), 0,
+						    "Failed to access child");
+			g_object_unref (G_OBJECT (child));
+			return NULL;
+		}
 	} else {
 		metabat = &info->bb.bat;
 		size_guess = dirent->size >> info->bb.shift;
@@ -865,6 +881,8 @@ gsf_infile_msole_class_init (GObjectClass *gobject_class)
 	infile_class->name_by_index	= gsf_infile_msole_name_by_index;
 	infile_class->child_by_index	= gsf_infile_msole_child_by_index;
 	infile_class->child_by_name	= gsf_infile_msole_child_by_name;
+
+	parent_class = g_type_class_peek_parent (gobject_class);
 }
 
 GSF_CLASS (GsfInfileMSOle, gsf_infile_msole,
@@ -881,7 +899,7 @@ GSF_CLASS (GsfInfileMSOle, gsf_infile_msole,
  *
  * Returns : the new ole file handler
  **/
-GsfInfileMSOle *
+GsfInfile *
 gsf_infile_msole_new (GsfInput *source, GError **err)
 {
 	GsfInfileMSOle *ole;
@@ -890,19 +908,20 @@ gsf_infile_msole_new (GsfInput *source, GError **err)
 	g_return_val_if_fail (GSF_IS_INPUT (source), NULL);
 
 	ole = (GsfInfileMSOle *)g_object_new (GSF_INFILE_MSOLE_TYPE, NULL);
-	g_object_ref (G_OBJECT (source));
-	ole->input = source;
-	gsf_input_set_size (GSF_INPUT (ole), (gsf_off_t) 0);
+	ole->input = gsf_input_proxy_new (source);
+	gsf_input_set_size (GSF_INPUT (ole), 0);
 
 	calling_pos = gsf_input_tell (source);
 	if (ole_init_info (ole, err)) {
-		gsf_input_seek (source, calling_pos, G_SEEK_SET);
+		/* It's not clear to me why we do this.  And if this
+		   fails, there's really nothing we can do.  */
+		(void)gsf_input_seek (source, calling_pos, G_SEEK_SET);
 
 		g_object_unref (G_OBJECT (ole));
 		return NULL;
 	}
 
-	return ole;
+	return GSF_INFILE (ole);
 }
 
 /**
