@@ -31,6 +31,8 @@
 
 #define Z_BUFSIZE 0x100
 
+static GObjectClass *parent_class;
+
 struct _GsfInputGZip {
 	GsfInput input;
 
@@ -68,7 +70,7 @@ check_header (GsfInputGZip *input)
 	unsigned flags, len;
 
 	/* Check signature */
-	if (gsf_input_seek (input->source, (gsf_off_t) 0, G_SEEK_SET) ||
+	if (gsf_input_seek (input->source, 0, G_SEEK_SET) ||
 	    NULL == (data = gsf_input_read (input->source, 2 + 1 + 1 + 6, NULL)) ||
 	    0 != memcmp (data, signature, sizeof (signature)))
 		return TRUE;
@@ -86,8 +88,7 @@ check_header (GsfInputGZip *input)
 	gsf_input_set_size (GSF_INPUT (input),
 			    (gsf_off_t) GSF_LE_GET_GUINT32 (data));
 
-	if (gsf_input_seek (input->source, (gsf_off_t) 2 + 1 + 1 + 6,
-			    G_SEEK_SET))
+	if (gsf_input_seek (input->source, 2 + 1 + 1 + 6, G_SEEK_SET))
 		return TRUE;
 
 	if (flags & GZIP_EXTRA_FIELD) {
@@ -184,7 +185,6 @@ gsf_input_gzip_new (GsfInput *source, GError **err)
 static void
 gsf_input_gzip_finalize (GObject *obj)
 {
-	GObjectClass *parent_class;
 	GsfInputGZip *gzip = (GsfInputGZip *)obj;
 
 	if (gzip->source != NULL) {
@@ -197,9 +197,7 @@ gsf_input_gzip_finalize (GObject *obj)
 	if (gzip->stream.state != NULL)
 		inflateEnd (&(gzip->stream));
 
-	parent_class = g_type_class_peek (GSF_INPUT_TYPE);
-	if (parent_class && parent_class->finalize)
-		parent_class->finalize (obj);
+	parent_class->finalize (obj);
 }
 
 static GsfInput *
@@ -334,8 +332,9 @@ gsf_input_gzip_class_init (GObjectClass *gobject_class)
 	input_class->Dup	= gsf_input_gzip_dup;
 	input_class->Read	= gsf_input_gzip_read;
 	input_class->Seek	= gsf_input_gzip_seek;
+
+	parent_class = g_type_class_peek_parent (gobject_class);
 }
 
 GSF_CLASS (GsfInputGZip, gsf_input_gzip,
 	   gsf_input_gzip_class_init, gsf_input_gzip_init, GSF_INPUT_TYPE)
-
