@@ -55,6 +55,7 @@ typedef enum {
 	GSF_XML_SHARED_CONTENT
 } GsfXMLContent;
 
+typedef gboolean (*GsfXMLInUnknownFunc) (GsfXMLIn *state, xmlChar const *elem, xmlChar const **attrs);
 struct _GsfXMLIn {
 	GsfXMLInDoc  const *doc;	/* init before parsing */
 
@@ -80,7 +81,7 @@ struct _GsfXMLInNode {
 	GSList *groups;
 
 	unsigned	has_content;
-	gboolean	allow_unknown;
+	gboolean	deprecated_unused_allow_unknown; /* remains here for binary compat */
 	gboolean	check_children_for_ns;
 
 	void (*start) (GsfXMLIn *state, xmlChar const **attrs);
@@ -103,10 +104,10 @@ struct _GsfXMLInNS {
 { uri, id}
 
 #define GSF_XML_IN_NODE_FULL(parent_id, id, ns, name, has_content, 	\
-			     allow_unknown, check_ns, start, end, user)	\
+			     deprecated_unused_allow_unknown, check_ns, start, end, user)	\
 {									\
 	#id, ns, name, #parent_id, FALSE, NULL,				\
-	has_content, allow_unknown, check_ns, start, end, { user } 	\
+	has_content, deprecated_unused_allow_unknown, check_ns, start, end, { user } 	\
 }
 
 #define GSF_XML_IN_NODE(parent_id, id, ns, name, has_content, start, end) \
@@ -115,6 +116,10 @@ struct _GsfXMLInNS {
 
 GsfXMLInDoc *gsf_xml_in_doc_new	 (GsfXMLInNode *root, GsfXMLInNS *ns);
 void	     gsf_xml_in_doc_free (GsfXMLInDoc *doc);
+void	     gsf_xml_in_doc_extend (GsfXMLInDoc  *doc,
+				    GsfXMLInNode *nodes);
+void	     gsf_xml_in_doc_set_unknown_handler (GsfXMLInDoc *doc,
+						 GsfXMLInUnknownFunc handler);
 
 gboolean    gsf_xml_in_parse	 (GsfXMLIn *state, GsfInput *input);
 char const *gsf_xml_in_check_ns	 (GsfXMLIn const *state, char const *str,
@@ -134,8 +139,9 @@ typedef struct _GsfXMLOut	GsfXMLOut;
 GType gsf_xml_out_get_type (void);
 GsfXMLOut *gsf_xml_out_new (GsfOutput *output);
 
+void	    gsf_xml_out_set_doc_type	(GsfXMLOut *xml, char const *type);
 void	    gsf_xml_out_start_element	(GsfXMLOut *xml, char const *id);
-char const *gsf_xml_out_end_element	(GsfXMLOut *xml);
+const char *gsf_xml_out_end_element	(GsfXMLOut *xml);
 
 void gsf_xml_out_simple_element		(GsfXMLOut *xml, char const *id,
 					 char const *content);
