@@ -2,12 +2,12 @@
 /*
  * gsf-msole-utils.c: 
  *
- * Copyright (C) 2002 Jody Goldberg (jody@gnome.org)
- * Copyright (C) 2002 Dom Lachowicz (cinamod@hotmail.com)
+ * Copyright (C) 2002-2004 Jody Goldberg (jody@gnome.org)
+ * Copyright (C) 2002-2003 Dom Lachowicz (cinamod@hotmail.com)
  * excel_iconv* family of functions (C) 2001 by Vlad Harchev <hvv@hippo.ru>
  *
  * This program is free software; you can redistribute it and/or
- * modify it under the terms of version 2 of the GNU General Public
+ * modify it under the terms of version 2.1 of the GNU Lesser General Public
  * License as published by the Free Software Foundation.
  *
  * This program is distributed in the hope that it will be useful,
@@ -15,7 +15,7 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
+ * You should have received a copy of the GNU Lesser General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  * USA
@@ -27,6 +27,7 @@
 #include <gsf/gsf-output.h>
 #include <gsf/gsf-utils.h>
 #include <gsf/gsf-timestamp.h>
+#include <gsf/gsf-meta-names.h>
 #include <stdio.h>
 
 #include <locale.h>
@@ -41,16 +42,28 @@
 #define d(code)
 #endif
 
+/*
+ * The Format Identifier for Summary Information
+ * F29F85E0-4FF9-1068-AB91-08002B27B3D9
+ */
 static guint8 const component_guid [] = {
 	0xe0, 0x85, 0x9f, 0xf2, 0xf9, 0x4f, 0x68, 0x10,
 	0xab, 0x91, 0x08, 0x00, 0x2b, 0x27, 0xb3, 0xd9
 };
 
+/*
+ * The Format Identifier for Document Summary Information
+ * D5CDD502-2E9C-101B-9397-08002B2CF9AE
+ */
 static guint8 const document_guid [] = {
 	0x02, 0xd5, 0xcd, 0xd5, 0x9c, 0x2e, 0x1b, 0x10,
 	0x93, 0x97, 0x08, 0x00, 0x2b, 0x2c, 0xf9, 0xae
 };
 
+/*
+ * The Format Identifier for User-Defined Properties
+ * D5CDD505-2E9C-101B-9397-08002B2CF9AE
+ */
 static guint8 const user_guid [] = {
 	0x05, 0xd5, 0xcd, 0xd5, 0x9c, 0x2e, 0x1b, 0x10,
 	0x93, 0x97, 0x08, 0x00, 0x2b, 0x2c, 0xf9, 0xae
@@ -107,7 +120,8 @@ typedef enum {
 } GsfMSOleVariantType;
 
 typedef struct {
-	char const *name;
+	char const *ms_name;
+	char const *gsf_name;
 	guint32	    id;
 	GsfMSOleVariantType prefered_type;
 } GsfMSOleMetaDataPropMap;
@@ -126,56 +140,69 @@ typedef struct {
 	GHashTable *dict;
 } GsfMSOleMetaDataSection;
 
+/*
+ * DocumentSummaryInformation properties
+ */
 static GsfMSOleMetaDataPropMap const document_props[] = {
-	{ "Category",		2,	VT_LPSTR },
-	{ "PresentationFormat",	3,	VT_LPSTR },
-	{ "NumBytes",		4,	VT_I4 },
-	{ "NumLines",		5,	VT_I4 },
-	{ "NumParagraphs",	6,	VT_I4 },
-	{ "NumSlides",		7,	VT_I4 },
-	{ "NumNotes",		8,	VT_I4 },
-	{ "NumHiddenSlides",	9,	VT_I4 },
-	{ "NumMMClips",		10,	VT_I4 },
-	{ "Scale",		11,	VT_BOOL },
-	{ "HeadingPairs",	12,	VT_VECTOR | VT_VARIANT },
-	{ "DocumentParts",	13,	VT_VECTOR | VT_LPSTR },
-	{ "Manager",		14,	VT_LPSTR },
-	{ "Company",		15,	VT_LPSTR },
-	{ "LinksDirty",		16,	VT_BOOL }
+	{ "Category",		GSF_META_NAME_CATEGORY,            2,	VT_LPSTR },
+	{ "PresentationFormat",	GSF_META_NAME_PRESENTATION_FORMAT, 3,	VT_LPSTR },
+	{ "NumBytes",		GSF_META_NAME_BYTE_COUNT,          4,	VT_I4 },
+	{ "NumLines",		GSF_META_NAME_LINE_COUNT,          5,	VT_I4 },
+	{ "NumParagraphs",	GSF_META_NAME_PARAGRAPH_COUNT,     6,	VT_I4 },
+	{ "NumSlides",		GSF_META_NAME_SLIDE_COUNT,         7,	VT_I4 },
+	{ "NumNotes",		GSF_META_NAME_NOTE_COUNT,          8,	VT_I4 },
+	{ "NumHiddenSlides",	GSF_META_NAME_HIDDEN_SLIDE_COUNT,  9,	VT_I4 },
+	{ "NumMMClips",		GSF_META_NAME_MM_CLIP_COUNT,       10,	VT_I4 },
+	{ "Scale",		GSF_META_NAME_SCALE,               11,	VT_BOOL },
+	{ "HeadingPairs",	GSF_META_NAME_HEADING_PAIRS,       12,	VT_VECTOR | VT_VARIANT },
+	{ "DocumentParts",	GSF_META_NAME_DOCUMENT_PARTS,      13,	VT_VECTOR | VT_LPSTR },
+	{ "Manager",		GSF_META_NAME_MANAGER,             14,	VT_LPSTR },
+	{ "Company",		GSF_META_NAME_COMPANY,             15,	VT_LPSTR },
+	{ "LinksDirty",		GSF_META_NAME_LINKS_DIRTY,         16,	VT_BOOL },
+	{ "DocSumInfo_17",      GSF_META_NAME_USER_DEFINED_1,      17,	VT_UNKNOWN },
+	{ "DocSumInfo_18",      GSF_META_NAME_USER_DEFINED_2,      18,	VT_UNKNOWN },
+	{ "DocSumInfo_19",      GSF_META_NAME_USER_DEFINED_3,      19,	VT_BOOL },
+	{ "DocSumInfo_20",      GSF_META_NAME_USER_DEFINED_4,      20,	VT_UNKNOWN },
+	{ "DocSumInfo_21",      GSF_META_NAME_USER_DEFINED_5,      21,	VT_UNKNOWN },
+	{ "DocSumInfo_22",      GSF_META_NAME_USER_DEFINED_6,      22,	VT_BOOL },
+	{ "DocSumInfo_23",      GSF_META_NAME_USER_DEFINED_7,      23,	VT_I4 }
 };
 
+/*
+ * SummaryInformation properties
+ */
 static GsfMSOleMetaDataPropMap const component_props[] = {
-	{ "Title",		2,	VT_LPSTR },
-	{ "Subject",		3,	VT_LPSTR },
-	{ "Author",		4,	VT_LPSTR },
-	{ "Keywords",		5,	VT_LPSTR },
-	{ "Comments",		6,	VT_LPSTR },
-	{ "Template",		7,	VT_LPSTR },
-	{ "LastSavedBy",	8,	VT_LPSTR },
-	{ "RevisionNumber",	9,	VT_LPSTR },
-	{ "TotalEditingTime",	10,	VT_FILETIME },
-	{ "LastPrinted",	11,	VT_FILETIME },
-	{ "CreateTime",		12,	VT_FILETIME },
-	{ "LastSavedTime",	13,	VT_FILETIME },
-	{ "NumPages",		14,	VT_I4 },
-	{ "NumWords",		15,	VT_I4 },
-	{ "NumCharacters",	16,	VT_I4 },
-	{ "Thumbnail",		17,	VT_CF },
-	{ "AppName",		18,	VT_LPSTR },
-	{ "Security",		19,	VT_I4 }
+	{ "Title",		GSF_META_NAME_TITLE,		2,	VT_LPSTR },
+	{ "Subject",		GSF_META_NAME_SUBJECT,		3,	VT_LPSTR },
+	{ "Author",		GSF_META_NAME_CREATOR,		4,	VT_LPSTR },
+	{ "Keywords",		GSF_META_NAME_KEYWORDS,		5,	VT_LPSTR },
+	{ "Comments",		GSF_META_NAME_DESCRIPTION,	6,	VT_LPSTR },
+	{ "Template",		GSF_META_NAME_TEMPLATE,		7,	VT_LPSTR },
+	{ "LastSavedBy",	GSF_META_NAME_LAST_SAVED_BY,	8,	VT_LPSTR },
+	{ "RevisionNumber",	GSF_META_NAME_REVISION_COUNT,	9,	VT_LPSTR },
+	{ "TotalEditingTime",	GSF_META_NAME_EDITING_DURATION,	10,	VT_FILETIME },
+	{ "LastPrinted",	GSF_META_NAME_LAST_PRINTED,	11,	VT_FILETIME },
+	{ "CreateTime",		GSF_META_NAME_DATE_CREATED,	12,	VT_FILETIME },
+	{ "LastSavedTime",	GSF_META_NAME_DATE_MODIFIED,	13,	VT_FILETIME },
+	{ "NumPages",		GSF_META_NAME_PAGE_COUNT,	14,	VT_I4 },
+	{ "NumWords",		GSF_META_NAME_WORD_COUNT,	15,	VT_I4 },
+	{ "NumCharacters",	GSF_META_NAME_CHARACTER_COUNT,	16,	VT_I4 },
+	{ "Thumbnail",		GSF_META_NAME_THUMBNAIL,	17,	VT_CF },
+	{ "AppName",		GSF_META_NAME_GENERATOR,	18,	VT_LPSTR },
+	{ "Security",		GSF_META_NAME_SECURITY,		19,	VT_I4 }
 };
 
 static GsfMSOleMetaDataPropMap const common_props[] = {
-	{ "Dictionary",		0,	0, /* magic */},
-	{ "CodePage",		1,	VT_UI2 },
-	{ "LOCALE_SYSTEM_DEFAULT",	0x80000000,	VT_UI4},
-	{ "CASE_SENSITIVE",		0x80000003,	VT_UI4},
+	{ "Dictionary",		  GSF_META_NAME_DICTIONARY,            0,	         0, /* magic */},
+	{ "CodePage",		  GSF_META_NAME_LANGUAGE,              1,	         VT_UI2 },
+	{ "LOCALE_SYSTEM_DEFAULT",GSF_META_NAME_LOCALE_SYSTEM_DEFAULT, 0x80000000,	 VT_UI4},
+	{ "CASE_SENSITIVE",	  GSF_META_NAME_CASE_SENSITIVE,        0x80000003,VT_UI4},
 };
 
 static char const *
 msole_prop_id_to_gsf (GsfMSOleMetaDataSection *section, guint32 id)
 {
-	char const * res = NULL;
+	char const *res = NULL;
 	GsfMSOleMetaDataPropMap const *map = NULL;
 	unsigned i = 0;
 
@@ -202,16 +229,16 @@ msole_prop_id_to_gsf (GsfMSOleMetaDataSection *section, guint32 id)
 	}
 	while (i-- > 0)
 		if (map[i].id == id) {
-			d (printf (map[i].name););
-			return map[i].name;
+			d (printf (map[i].gsf_name););
+			return map[i].gsf_name;
 		}
 
 	map = common_props;
 	i = G_N_ELEMENTS (common_props);
 	while (i-- > 0)
 		if (map[i].id == id) {
-			d (printf (map[i].name););
-			return map[i].name;
+			d (printf (map[i].gsf_name););
+			return map[i].gsf_name;
 		}
 
 	d (printf ("_UNKNOWN_(0x%x %d)", id, id););
@@ -226,6 +253,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 	GValue *res;
 	char *str;
 	guint32 len;
+	gsize gslen;
 	gboolean const is_vector = type & VT_VECTOR;
 
 	g_return_val_if_fail (!(type & (unsigned)(~0x1fff)), NULL); /* not valid in a prop set */
@@ -233,6 +261,12 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 	type &= 0xfff;
 
 	if (is_vector) {
+		/*
+		 *  A vector is basically an array.  If the type associated with
+		 *  it is a variant, then each element can have a different
+		 *  variant type.  Otherwise, each element has the same variant
+		 *  type associated with the vector.
+		 */
 		unsigned i, n;
 
 		g_return_val_if_fail (*data + 4 <= data_end, NULL);
@@ -243,8 +277,15 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		d (printf (" array with %d elem\n", n);
 		   gsf_mem_dump (*data, (unsigned)(data_end - *data)););
 		for (i = 0 ; i < n ; i++) {
+			GValue *v;
 			d (printf ("\t[%d] ", i););
-			msole_prop_parse (section, type, data, data_end);
+			v = msole_prop_parse (section, type, data, data_end);
+			if (v) {
+				/* FIXME: do something with it.  */
+				if (G_IS_VALUE (v))
+					g_value_unset (v);
+				g_free (v);
+			}
 		}
 		return NULL;
 	}
@@ -252,14 +293,20 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 	res = g_new0 (GValue, 1);
 	switch (type) {
 	case VT_EMPTY :		 d (puts ("VT_EMPTY"););
+		/*
+		 * A property with a type indicator of VT_EMPTY has no data
+		 * associated with it; that is, the size of the value is zero.
+		 */
 		/* value::unset == empty */
 		break;
 
 	case VT_NULL :		 d (puts ("VT_NULL"););
+		/* This is like a pointer to NULL */
 		/* value::unset == null too :-) do we need to distinguish ? */
 		break;
 
 	case VT_I2 :		 d (puts ("VT_I2"););
+		/* 2-byte signed integer */
 		g_return_val_if_fail (*data + 2 <= data_end, NULL);
 		g_value_init (res, G_TYPE_INT);
 		g_value_set_int	(res, GSF_LE_GET_GINT16 (*data));
@@ -267,6 +314,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		break;
 
 	case VT_I4 :		 d (puts ("VT_I4"););
+		/* 4-byte signed integer */
 		g_return_val_if_fail (*data + 4 <= data_end, NULL);
 		g_value_init (res, G_TYPE_INT);
 		g_value_set_int	(res, GSF_LE_GET_GINT32 (*data));
@@ -274,6 +322,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		break;
 
 	case VT_R4 :		 d (puts ("VT_R4"););
+		/* 32-bit IEEE floating-point value */
 		g_return_val_if_fail (*data + 4 <= data_end, NULL);
 		g_value_init (res, G_TYPE_FLOAT);
 		g_value_set_float (res, GSF_LE_GET_FLOAT (*data));
@@ -281,6 +330,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		break;
 
 	case VT_R8 :		 d (puts ("VT_R8"););
+		/* 64-bit IEEE floating-point value */
 		g_return_val_if_fail (*data + 8 <= data_end, NULL);
 		g_value_init (res, G_TYPE_DOUBLE);
 		g_value_set_double (res, GSF_LE_GET_DOUBLE (*data));
@@ -288,18 +338,33 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		break;
 
 	case VT_CY :		 d (puts ("VT_CY"););
+		/* 8-byte two's complement integer (scaled by 10,000) */
+#warning TODO
 		break;
 
 	case VT_DATE :		 d (puts ("VT_DATE"););
+		/* 
+		 * 64-bit floating-point number representing the number of days
+		 * (not seconds) since December 31, 1899.
+		 */
+#warning TODO
 		break;
 
 	case VT_BSTR :		 d (puts ("VT_BSTR"););
+		/*
+		 * Pointer to null-terminated Unicode string; the string is pre-
+		 * ceeded by a DWORD representing the byte count of the number
+		 * of bytes in the string (including the  terminating null).
+		 */
+#warning TODO
 		break;
 
 	case VT_DISPATCH :	 d (puts ("VT_DISPATCH"););
+#warning TODO
 		break;
 
 	case VT_BOOL :		 d (puts ("VT_BOOL"););
+		/* A boolean (WORD) value containg 0 (false) or -1 (true). */
 		g_return_val_if_fail (*data + 1 <= data_end, NULL);
 		g_value_init (res, G_TYPE_BOOLEAN);
 		g_value_set_boolean (res, **data ? TRUE : FALSE);
@@ -307,12 +372,18 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		break;
 
 	case VT_VARIANT :	 d (printf ("VT_VARIANT containing a "););
+		/*
+		 * A type indicator (a DWORD) followed by the corresponding
+		 *  value.  VT_VARIANT is only used in conjunction with
+		 *  VT_VECTOR.
+		 */
 		g_free (res);
 		type = GSF_LE_GET_GUINT32 (*data);
 		*data += 4;
 		return msole_prop_parse (section, type, data, data_end);
 
 	case VT_UI1 :		 d (puts ("VT_UI1"););
+		/* 1-byte unsigned integer */
 		g_return_val_if_fail (*data + 1 <= data_end, NULL);
 		g_value_init (res, G_TYPE_UCHAR);
 		g_value_set_uchar (res, (guchar)(**data));
@@ -320,6 +391,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		break;
 
 	case VT_UI2 :		 d (puts ("VT_UI2"););
+		/* 2-byte unsigned integer */
 		g_return_val_if_fail (*data + 2 <= data_end, NULL);
 		g_value_init (res, G_TYPE_UINT);
 		g_value_set_uint (res, GSF_LE_GET_GUINT16 (*data));
@@ -327,6 +399,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		break;
 
 	case VT_UI4 :		 d (puts ("VT_UI4"););
+		/* 4-type unsigned integer */
 		g_return_val_if_fail (*data + 4 <= data_end, NULL);
 		g_value_init (res, G_TYPE_UINT);
 		*data += 4;
@@ -334,28 +407,39 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		break;
 
 	case VT_I8 :		 d (puts ("VT_I8"););
+		/* 8-byte signed integer */
 		g_return_val_if_fail (*data + 8 <= data_end, NULL);
 		g_value_init (res, G_TYPE_INT64);
 		*data += 8;
 		break;
 
 	case VT_UI8 :		 d (puts ("VT_UI8"););
+		/* 8-byte unsigned integer */
 		g_return_val_if_fail (*data + 8 <= data_end, NULL);
 		g_value_init (res, G_TYPE_UINT64);
 		*data += 8;
 		break;
 
 	case VT_LPSTR :		 d (puts ("VT_LPSTR"););
+		/* 
+		 * This is the representation of many strings.  It is stored in
+		 * the same representation as VT_BSTR.  Note that the serialized
+		 * representation of VP_LPSTR has a preceding byte count, wheras
+		 * the in-memory representation does not.
+		 */
 		/* be anal and safe */
 		g_return_val_if_fail (*data + 4 <= data_end, NULL);
 
 		len = GSF_LE_GET_GUINT32 (*data);
 
-		g_return_val_if_fail (*data + 4 + len <= data_end, NULL);
+		g_return_val_if_fail (len < 0x10000, NULL);
+		g_return_val_if_fail (*data + 4 + len*section->char_size <= data_end, NULL);
 
+		gslen = 0;
 		str = g_convert_with_iconv (*data + 4,
 			len * section->char_size,
-			section->iconv_handle, &len, NULL, NULL);
+			section->iconv_handle, &gslen, NULL, NULL);
+		len = (guint32)gslen;
 
 		g_value_init (res, G_TYPE_STRING);
 		g_value_set_string (res, str);
@@ -364,15 +448,23 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		break;
 
 	case VT_LPWSTR : d (puts ("VT_LPWSTR"););
+		/*
+		 * A counted and null-terminated Unicode string; a DWORD character
+		 * count (where the count includes the terminating null) followed
+		 * by that many Unicode (16-bit) characters.  Note that the count
+		 * is character count, not byte count.
+		 */
 		/* be anal and safe */
 		g_return_val_if_fail (*data + 4 <= data_end, NULL);
 
 		len = GSF_LE_GET_GUINT32 (*data);
 
+		g_return_val_if_fail (len < 0x10000, NULL);
 		g_return_val_if_fail (*data + 4 + len <= data_end, NULL);
 
 		str = g_convert (*data + 4, len*2,
-				 "UTF-8", "UTF-16LE", &len, NULL, NULL);
+				 "UTF-8", "UTF-16LE", &gslen, NULL, NULL);
+		len = (guint32)gslen;
 
 		g_value_init (res, G_TYPE_STRING);
 		g_value_set_string (res, str);
@@ -381,6 +473,7 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		break;
 
 	case VT_FILETIME :	 d (puts ("VT_FILETIME"););
+		/* 64-bit FILETIME structure, as defined by Win32. */
 		g_return_val_if_fail (*data + 8 <= data_end, NULL);
 		g_value_init (res, GSF_TIMESTAMP_TYPE);
 	{
@@ -389,34 +482,116 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		GsfTimestamp ts;
 
 		ft /= 10000000; /* convert to seconds */
-		ft -= 11644473600; /* move to Jan 1 1970 */
+#ifdef _MSC_VER
+		ft -= 11644473600i64; /* move to Jan 1 1970 */
+#else
+		ft -= 11644473600ULL; /* move to Jan 1 1970 */
+#endif
 		ts.timet = (time_t)ft;
-		g_value_set_timestamp (res, &ts);
+		gsf_value_set_timestamp (res, &ts);
 		*data += 8;
 		break;
 	}
 	case VT_BLOB :		 d (puts ("VT_BLOB"););
+		/*
+		 * A DWORD count of bytes, followed by that many bytes of data.
+		 * The byte count does not include the four bytes for the length
+		 * of the count itself:  An empty blob would have a count of
+		 * zero, followed by zero bytes.  Thus the serialized represen-
+		 * tation of a VT_BLOB is similar to that of a VT_BSTR but does
+		 * not guarantee a null byte at the end of the data.
+		 */
+#warning TODO
+		g_free (res);
+		res = NULL;
 		break;
+
 	case VT_STREAM :	 d (puts ("VT_STREAM"););
+		/*
+		 * Indicates the value is stored in a stream that is sibling to
+		 * the CONTENTS stream.  Following this type indicator is data
+		 * in the format of a serialized VT_LPSTR, which names the stream
+		 * containing the data.
+		 */
+#warning TODO
+		g_free (res);
+		res = NULL;
 		break;
+
 	case VT_STORAGE :	 d (puts ("VT_STORAGE"););
+		/*
+		 * Indicates the value is stored in an IStorage that is sibling
+		 * to the CONTENTS stream.  Following this type indicator is data
+		 * in the format of a serialized VT_LPSTR, which names the
+		 * IStorage containing the data.
+		 */
+#warning TODO
+		g_free (res);
+		res = NULL;
 		break;
+
 	case VT_STREAMED_OBJECT: d (puts ("VT_STREAMED_OBJECT"););
+		/*
+		 * Same as VT_STREAM, but indicates that the stream contains a
+		 * serialized object, which is a class ID followed by initiali-
+		 * zation data for the class.
+		 */
+#warning TODO
+		g_free (res);
+		res = NULL;
 		break;
+
 	case VT_STORED_OBJECT :	 d (puts ("VT_STORED_OBJECT"););
+		/*
+		 * Same as VT_STORAGE, but indicates that the designated IStorage
+		 * contains a loadable object.
+		 */
+#warning TODO
+		g_free (res);
+		res = NULL;
 		break;
+
 	case VT_BLOB_OBJECT :	 d (puts ("VT_BLOB_OBJECT"););
+		/*
+		 * Contains a serialized object in the same representation as
+		 * would appear in a VT_STREAMED_OBJECT.  That is, following the
+		 * VT_BLOB_OBJECT tag is a DWORD byte count of the remaining data
+		 * (where the byte count does not include the size of itself)
+		 * which is in the format of a class ID followed by initialization
+		 * data for that class
+		 */
+#warning TODO
+		g_free (res);
+		res = NULL;
 		break;
+
 	case VT_CF :		 d (puts ("VT_CF"););
+		/*
+		 * a BLOB containing a clipboard format identifier followed by
+		 * the data in that format.  That is, following the VT_CF tag is
+		 * data in the format of a VT_BLOB: a CWORD count of bytes,
+		 * followed by that many bytes of data in the format of a packed
+		 * VTCFREP, followed immediately by an array of bytes as appropriate
+		 * for data in the clipboard format.
+		 */
+#warning TODO
+		g_free (res);
+		res = NULL;
 		break;
+
 	case VT_CLSID :		 d (puts ("VT_CLSID"););
+		/* A class ID (or other GUID) */
 		*data += 16;
+		g_free (res);
+		res = NULL;
 		break;
 
 	case VT_ERROR :
+		/* A DWORD containing a status code. */
 	case VT_UNKNOWN :
 	case VT_DECIMAL :
 	case VT_I1 :
+		/* 1-byte signed integer */
 	case VT_INT :
 	case VT_UINT :
 	case VT_VOID :
@@ -435,19 +610,23 @@ msole_prop_parse (GsfMSOleMetaDataSection *section,
 		g_warning ("Unknown property type %d (0x%x)", type, type);
 		g_free (res);
 		res = NULL;
-	};
+	}
 
-	d ( if (res != NULL) {
-		char *val = g_strdup_value_contents (res);
-		printf ("%s\n", val);
-		g_free (val);
-	} else
+	if (res != NULL && G_IS_VALUE (res)) {
+		d ( {
+			char *val = g_strdup_value_contents (res);
+			printf ("%s\n", val);
+			g_free (val);
+		});
+	} else {
+#if 0 /* enable after release */
 		puts ("<unparsed>\n");
-	);
+#endif
+	}
 	return res;
 }
 
-static GValue *
+static GsfDocProp *
 msole_prop_read (GsfInput *in,
 		 GsfMSOleMetaDataSection *section,
 		 GsfMSOleMetaDataProp    *props,
@@ -456,13 +635,14 @@ msole_prop_read (GsfInput *in,
 	guint32 type;
 	guint8 const *data;
 	/* TODO : why size-4 ? I must be missing something */
-	unsigned size = ((i+1) >= section->num_props)
+	gsf_off_t size = ((i+1) >= section->num_props)
 		? section->size-4 : props[i+1].offset;
-	char const *prop_name;
+	GsfDocProp	*prop = NULL;
 
 	g_return_val_if_fail (i < section->num_props, NULL);
 	g_return_val_if_fail (size >= props[i].offset + 4, NULL);
 
+	prop = g_new (GsfDocProp, 1);
 	size -= props[i].offset; /* includes the type id */
 	if (gsf_input_seek (in, section->offset+props[i].offset, G_SEEK_SET) ||
 	    NULL == (data = gsf_input_read (in, size, NULL))) {
@@ -476,6 +656,7 @@ msole_prop_read (GsfInput *in,
 	/* dictionary is magic */
 	if (props[i].id == 0) {
 		guint32 len, id, i, n;
+		gsize gslen;
 		char *name;
 		guint8 const *start = data;
 
@@ -493,9 +674,11 @@ msole_prop_read (GsfInput *in,
 
 			g_return_val_if_fail (len < 0x10000, NULL);
 
+			gslen = 0;
 			name = g_convert_with_iconv (data + 8,
 				len * section->char_size,
-				section->iconv_handle, &len, NULL, NULL);
+				section->iconv_handle, &gslen, NULL, NULL);
+			len = (guint32)gslen;
 			data += 8 + len;
 
 			d (printf ("\t%u == %s\n", id, name););
@@ -515,10 +698,12 @@ msole_prop_read (GsfInput *in,
 	}
 
 	d (printf ("%u) ", i););
-	prop_name = msole_prop_id_to_gsf (section, props[i].id);
+	prop->name = msole_prop_id_to_gsf (section, props[i].id);
 
-	d (printf (" @ %x %x = ", (unsigned)props[i].offset, size););
-	return msole_prop_parse (section, type, &data, data + size);
+	d (printf (" @ %x %x = ", (unsigned)props[i].offset, (unsigned)size););
+	prop->val = msole_prop_parse (section, type, &data, data + size);
+
+	return prop;
 }
 
 static int
@@ -529,8 +714,38 @@ msole_prop_cmp (gconstpointer a, gconstpointer b)
 	return prop_a->offset - prop_b->offset;
 }
 
+static void
+msole_prop_store (GsfDocMetaData *ret_object, GsfDocProp *prop)
+{
+	if (NULL != prop && NULL != prop->val) {
+		gsf_doc_meta_data_set_prop (ret_object,
+			prop->name, prop->val);
+		if (G_IS_VALUE (prop->val))
+			g_value_unset (prop->val);
+		g_free (prop->val);
+	}
+}
+
 gboolean
 gsf_msole_metadata_read (GsfInput *in, GError **err)
+{
+	GsfDocMetaData *meta = gsf_msole_metadata_read_real (in, err);
+	g_object_unref (meta);
+	return meta != NULL;
+}
+
+/**
+ * gsf_msole_metadata_read_real:
+ * @in : #GsfInput
+ * @err : #GError
+ *
+ * Parse an OLE2 property set out of @in and return a #GsfDocMetaData
+ * with the content.  For now this is called _real, to avoid inadvertent
+ * leakage for callers expecting the old bool return.  Callers is responsible
+ * for releasing the result.
+ **/
+GsfDocMetaData*
+gsf_msole_metadata_read_real (GsfInput *in, GError **err)
 {
 	guint8 const *data = gsf_input_read (in, 28, NULL);
 	guint16 version;
@@ -538,18 +753,28 @@ gsf_msole_metadata_read (GsfInput *in, GError **err)
 	unsigned i, j;
 	GsfMSOleMetaDataSection *sections;
 	GsfMSOleMetaDataProp *props;
-
+	GsfDocMetaData	*ret_object = NULL;
+	GsfDocProp	*prop;
+	
 	if (NULL == data) {
 		if (err != NULL)
 			*err = g_error_new (gsf_input_error (), 0,
 					    "Unable to read MS property stream header");
-		return FALSE;
+		return ret_object;
 	}
 
-	/* NOTE : high word is the os, low word is the os version
-	 * 0 = win16
-	 * 1 = mac
-	 * 2 = win32
+	/*
+	 * Validate the Property Set Header.
+	 * Format (bytes) :
+	 *   00 - 01	Byte order		0xfffe
+	 *   02 - 03	Format			0
+	 *   04 - 05	OS Version		high word is the OS
+	 *   06 - 07				low  word is the OS version
+	 *					  0 = win16
+	 *					  1 = mac
+	 *					  2 = win32
+	 *   08 - 23	Class Identifier	Usually Format ID
+	 *   24 - 27	Section count		Should be at least 1
 	 */
 	os = GSF_LE_GET_GUINT16 (data + 6);
 	version = GSF_LE_GET_GUINT16 (data + 2);
@@ -562,18 +787,27 @@ gsf_msole_metadata_read (GsfInput *in, GError **err)
 		if (err != NULL)
 			*err = g_error_new (gsf_input_error (), 0,
 					    "Invalid MS property stream header");
-		return FALSE;
+		return ret_object;
 	}
 
 	/* extract the section info */
-	sections = g_alloca (sizeof (GsfMSOleMetaDataSection)* num_sections);
+	/*
+	 * The Format ID/Offset list follows.
+	 * Format:
+	 *   00 - 16	Section Name		Format ID
+	 *   16 - 19	Section Offset		The offset is the number of
+	 *					bytes from the start of the
+	 *					whole stream to where the
+	 *					section begins.
+	 */
+	sections = (GsfMSOleMetaDataSection *)g_alloca (sizeof (GsfMSOleMetaDataSection)* num_sections);
 	for (i = 0 ; i < num_sections ; i++) {
 		data = gsf_input_read (in, 20, NULL);
 		if (NULL == data) {
 			if (err != NULL)
 				*err = g_error_new (gsf_input_error (), 0,
 						    "Unable to read MS property stream header");
-			return FALSE;
+			return ret_object;
 		}
 		if (!memcmp (data, component_guid, sizeof (component_guid)))
 			sections [i].type = GSF_MSOLE_META_DATA_COMPONENT;
@@ -588,17 +822,28 @@ gsf_msole_metadata_read (GsfInput *in, GError **err)
 		}
 
 		sections [i].offset = GSF_LE_GET_GUINT32 (data + 16);
-#ifndef NO_DEBUG_OLE_PROPS
-		printf ("0x%x\n", (guint32)sections [i].offset);
-#endif
+		d (printf ("0x%x\n", (guint32)sections [i].offset););
 	}
+
+	/*
+	 * A section is the third part of the property set stream.
+	 * Format (bytes) :
+	 *   00 - 03	Section size	A byte count for the section (which is inclusive
+	 *				of the byte count itself and should always be a
+	 *				multiple of 4);
+	 *   04 - 07	Property count	A count of the number of properties
+	 *   08 - xx   			An array of 32-bit Property ID/Offset pairs
+	 *   yy - zz			An array of Property Type indicators/Value pairs
+	 */
+	ret_object = gsf_doc_meta_data_new ();
+
 	for (i = 0 ; i < num_sections ; i++) {
 		if (gsf_input_seek (in, sections[i].offset, G_SEEK_SET) ||
 		    NULL == (data = gsf_input_read (in, 8, NULL))) {
 			if (err != NULL)
 				*err = g_error_new (gsf_input_error (), 0,
 						    "Invalid MS property section");
-			return FALSE;
+			return ret_object;
 		}
 
 		sections[i].iconv_handle = (GIConv)-1;
@@ -608,6 +853,14 @@ gsf_msole_metadata_read (GsfInput *in, GError **err)
 		sections[i].num_props = GSF_LE_GET_GUINT32 (data + 4);
 		if (sections[i].num_props <= 0)
 			continue;
+
+		/*
+		 * Get and save all the Property ID/Offset pairs.
+		 * Format (bytes) :
+		 *   00 - 03	id	Property ID
+		 *   04 - 07	offset	The distance from the start of the section to the
+		 *			start of the Property Type/Value pair.
+		 */
 		props = g_new (GsfMSOleMetaDataProp, sections[i].num_props);
 		for (j = 0; j < sections[i].num_props; j++) {
 			if (NULL == (data = gsf_input_read (in, 8, NULL))) {
@@ -615,7 +868,7 @@ gsf_msole_metadata_read (GsfInput *in, GError **err)
 				if (err != NULL)
 					*err = g_error_new (gsf_input_error (), 0,
 							    "Invalid MS property section");
-				return FALSE;
+				return ret_object;
 			}
 
 			props [j].id = GSF_LE_GET_GUINT32 (data);
@@ -627,36 +880,56 @@ gsf_msole_metadata_read (GsfInput *in, GError **err)
 		       sizeof (GsfMSOleMetaDataProp),
 		       msole_prop_cmp);
 
+		/*
+		 * Find and process the code page.
+		 * Property ID 1 is reserved as an indicator of the code page.
+		 */
 		sections[i].iconv_handle = (GIConv)-1;
 		sections[i].char_size = 1;
 		for (j = 0; j < sections[i].num_props; j++) /* first codepage */
 			if (props[j].id == 1) {
-				GValue *v = msole_prop_read (in, sections+i, props, j);
-				if (v != NULL) {
-					int codepage = g_value_get_int (v);
-					sections[i].iconv_handle = gsf_msole_iconv_open_for_import (codepage);
-					if (codepage == 1200 || codepage == 1201)
-						sections[i].char_size = 2;
-					g_value_unset (v);
-					g_free (v) ;
+				prop = msole_prop_read (in, sections+i, props, j);
+				if (NULL != prop && NULL != prop->val) {
+					gsf_doc_meta_data_set_prop (ret_object,
+						prop->name, prop->val);
+
+					if (G_IS_VALUE (prop->val)) {
+						if (G_VALUE_HOLDS_INT (prop->val)) {
+							int codepage = g_value_get_int (prop->val);
+							sections[i].iconv_handle = gsf_msole_iconv_open_for_import (codepage);
+							if (codepage == 1200 || codepage == 1201)
+								sections[i].char_size = 2;
+						}
+						g_value_unset (prop->val);
+					}
+					g_free (prop->val);
 				}
 			}
 		if (sections[i].iconv_handle == (GIConv)-1)
 			sections[i].iconv_handle = gsf_msole_iconv_open_for_import (1252);
 
+		/*
+		 * Find and process the Property Set Dictionary
+		 * Property ID 0 is reserved as an indicator of the dictionary.
+		 * For User Defined Sections, Property ID 0 is NOT a dictionary.
+		 */
 		for (j = 0; j < sections[i].num_props; j++) /* then dictionary */
 			if (props[j].id == 0)
-				msole_prop_read (in, sections+i, props, j);
+				msole_prop_store (ret_object,
+					msole_prop_read (in, sections+i, props, j));
+
+		/* Process all the properties */
 		for (j = 0; j < sections[i].num_props; j++) /* the rest */
 			if (props[j].id > 1)
-				msole_prop_read (in, sections+i, props, j);
+				msole_prop_store (ret_object,
+					msole_prop_read (in, sections+i, props, j));
 
 		gsf_iconv_close (sections[i].iconv_handle);
 		g_free (props);
 		if (sections[i].dict != NULL)
 			g_hash_table_destroy (sections[i].dict);
 	}
-	return TRUE;
+	return ret_object;
 }
 
 gboolean
@@ -711,7 +984,7 @@ static GsfLanguageMapping const gsf_msole_language_ids[] = {
 	{ "-none-", 0x0400 }, /* none */
 	{ "af_ZA",  0x0436 }, /* Afrikaans */
 	{ "am",     0x045e }, /* Amharic */
-	{ "al_AL",  0x041c }, /* Albanian */
+	{ "sq_AL",  0x041c }, /* Albanian */
 	{ "ar_SA",  0x0401 }, /* Arabic (Saudi) */
 	{ "ar_IQ",  0x0801 }, /* Arabic (Iraq) */
 	{ "ar_EG",  0x0c01 }, /* Arabic (Egypt) */		
@@ -889,20 +1162,24 @@ static GsfLanguageMapping const gsf_msole_language_ids[] = {
 
 /**
  * gsf_msole_lid_for_language
+ * @lang :
  *
- * Return the LID (Language Identifier) for the input language.
+ * Returns the LID (Language Identifier) for the input language.
  * If lang is %null, return 0x0400 ("-none-"), and not 0x0000 ("no proofing")
- */
+ **/
 guint
-gsf_msole_lid_for_language (const char * lang)
+gsf_msole_lid_for_language (char const *lang)
 {
 	guint i = 0 ;
+	size_t len;
 
 	if (lang == NULL)
 		return 0x0400;   /* return -none- */
-	
+
+	/* Allow lang to match as a prefix (eg fr == fr_FR@euro) */
+	len = strlen (lang);
 	for (i = 0 ; i < G_N_ELEMENTS(gsf_msole_language_ids); i++)
-		if (!strcmp (lang, gsf_msole_language_ids[i].tag))
+		if (!strncmp (lang, gsf_msole_language_ids[i].tag, len))
 			return gsf_msole_language_ids[i].lid;
 	
 	return 0x0400 ;   /* return -none- */
@@ -914,7 +1191,7 @@ gsf_msole_lid_for_language (const char * lang)
  * Returns the xx_YY style string (can be just xx or xxx) for the given LID.
  * Return value must not be freed. If the LID is not found, is set to 0x0400,
  * or is set to 0x0000, will return "-none-"
- */
+ **/
 G_CONST_RETURN char *
 gsf_msole_language_for_lid (guint lid)
 {
@@ -931,9 +1208,9 @@ gsf_msole_language_for_lid (guint lid)
  * gsf_msole_locale_to_lid :
  *
  * Covert the the codepage into an applicable LID
- */
+ **/
 guint
-gsf_msole_codepage_to_lid (guint codepage)
+gsf_msole_codepage_to_lid (int codepage)
 {
 	switch (codepage) {
 	case 77:		/* MAC_CHARSET */
@@ -974,9 +1251,10 @@ gsf_msole_codepage_to_lid (guint codepage)
 
 /**
  * gsf_msole_lid_to_codepage
+ * @lid :
  *
- * Our best guess at the codepage for the given language id
- */
+ * Returns our best guess at the codepage for the given language id
+ **/
 guint
 gsf_msole_lid_to_codepage (guint lid)
 {
@@ -1000,6 +1278,8 @@ gsf_msole_lid_to_codepage (guint lid)
 			
 		case 0x0804:		/* Chinese (PRC) */
 			return 936;
+		default :
+			break;
 		}
 		break;
 	case 0x05:		/* Czech */
@@ -1034,6 +1314,8 @@ gsf_msole_lid_to_codepage (guint lid)
 			return 1361;
 		case 0x0412:		/* Korean */
 			return 949;
+		default :
+			break;
 		}
 		break;
 	case 0x13:		/* Dutch */
@@ -1058,6 +1340,8 @@ gsf_msole_lid_to_codepage (guint lid)
 			return 1251;
 		case 0x081a:		/* Serbian (Latin) */
 			return 1252;
+		default :
+			break;
 		}
 		break;
 	case 0x1b:		/* Slovak */
@@ -1096,6 +1380,8 @@ gsf_msole_lid_to_codepage (guint lid)
 		switch (lid) {
 		case 0x082c:		/* Azeri (Cyrillic) */
 			return 1251;
+		default :
+			break;
 		}
 		break;
 	case 0x2d:		/* Basque */
@@ -1118,6 +1404,8 @@ gsf_msole_lid_to_codepage (guint lid)
 		switch (lid) {
 		case 0x0843:		/* Uzbek (Cyrillic) */
 			return 1251;
+		default :
+			break;
 		}
 		break;
 	case 0x45:		/* Windows 2000: Bengali. This is Unicode only. */
@@ -1164,10 +1452,11 @@ gsf_msole_lid_to_codepage (guint lid)
 
 /**
  * gsf_msole_lid_to_codepage_str
+ * @lid :
  * 
  * Returns the Iconv codepage string for the given LID.
  * Return value must be g_free()'d
- */
+ **/
 gchar *
 gsf_msole_lid_to_codepage_str (guint lid)
 {
@@ -1183,8 +1472,8 @@ gsf_msole_lid_to_codepage_str (guint lid)
 /**
  * gsf_msole_iconv_win_codepage :
  *
- * Our best guess at the applicable windows code page based on an environment
- * variable or the current locale.
+ * Returns our best guess at the applicable windows code page based on an
+ * 	environment variable or the current locale.
  **/
 guint
 gsf_msole_iconv_win_codepage (void)
@@ -1212,34 +1501,50 @@ gsf_msole_iconv_win_codepage (void)
 
 /**
  * gsf_msole_iconv_open_codepage_for_import :
- * @to
+ * @to:
  * @codepage :
  *
- * Open an iconv converter for @codepage -> utf8.
+ * Returns an iconv converter for @codepage -> utf8.
  **/
 GIConv
-gsf_msole_iconv_open_codepage_for_import (const char * to, guint codepage)
+gsf_msole_iconv_open_codepage_for_import (char const *to, int codepage)
 {
 	GIConv iconv_handle;
-
 	g_return_val_if_fail (to != NULL, (GIConv)(-1));
 
-	if (codepage != 1200 && codepage != 1201) {
+	/* sometimes it is stored as signed short */
+	if (codepage == 65001 || codepage == -535) {
+		iconv_handle = g_iconv_open (to, "UTF-8");
+		if (iconv_handle != (GIConv)(-1))
+			return iconv_handle;
+	} else if (codepage != 1200 && codepage != 1201) {
 		char* src_charset = g_strdup_printf ("CP%d", codepage);
 		iconv_handle = g_iconv_open (to, src_charset);
 		g_free (src_charset);
 		if (iconv_handle != (GIConv)(-1))
 			return iconv_handle;
-		g_warning ("Unknown codepage %d", codepage);
-		return (GIConv)(-1);
+	} else {
+		char const *from = (codepage == 1200) ? "UTF-16LE" : "UTF-16BE";
+		iconv_handle = g_iconv_open (to, from);
+		if (iconv_handle != (GIConv)(-1))
+			return iconv_handle;
 	}
 
-	iconv_handle = g_iconv_open (to,
-		(codepage == 1200) ? "UTF-16LE" : "UTF-16BE");
-	if (iconv_handle != (GIConv)(-1))
-		return iconv_handle;
-	g_warning ("Unable to open an iconv handle from %s -> %s",
-		   (codepage == 1200) ? "UTF-16LE" : "UTF-16BE", to);
+	/* Try aliases.  */
+	if (codepage == 10000) {
+		/* gnu iconv.  */
+		iconv_handle = g_iconv_open (to, "MACROMAN");
+		if (iconv_handle != (GIConv)(-1))
+			return iconv_handle;
+
+		/* glibc.  */
+		iconv_handle = g_iconv_open (to, "MACINTOSH");
+		if (iconv_handle != (GIConv)(-1))
+			return iconv_handle;
+	}
+
+	g_warning ("Unable to open an iconv handle from codepage %d -> %s",
+		   codepage, to);
 	return (GIConv)(-1);
 }
 
@@ -1247,21 +1552,23 @@ gsf_msole_iconv_open_codepage_for_import (const char * to, guint codepage)
  * gsf_msole_iconv_open_for_import :
  * @codepage :
  *
- * Open an iconv converter for single byte encodings @codepage -> utf8.
- * Attempt to handle the semantics of a specification for multibyte encodings
- * since this is only supposed to be used for single bytes.
+ * Returns an iconv converter for single byte encodings @codepage -> utf8.
+ * 	Attempt to handle the semantics of a specification for multibyte encodings
+ * 	since this is only supposed to be used for single bytes.
  **/
 GIConv
-gsf_msole_iconv_open_for_import (guint codepage)
+gsf_msole_iconv_open_for_import (int codepage)
 {
 	return gsf_msole_iconv_open_codepage_for_import ("UTF-8", codepage);
 }
 
 /**
- * gsf_msole_iconv_open_for_export :
+ * gsf_msole_iconv_open_codepages_for_export :
+ * @codepage_to :
+ * @from :
  *
- * Open an iconv convert to go from utf8 -> to our best guess at a useful
- * windows codepage.
+ * Returns an iconv converter to go from utf8 -> to our best guess at a useful
+ * 	windows codepage.
  **/
 GIConv
 gsf_msole_iconv_open_codepages_for_export (guint codepage_to, char const *from)
@@ -1280,11 +1587,11 @@ gsf_msole_iconv_open_codepages_for_export (guint codepage_to, char const *from)
 }
 
 /**
- * gsf_msole_iconv_open_for_export :
- * @to
+ * gsf_msole_iconv_open_codepage_for_export :
+ * @codepage_to:
  *
- * Open an iconv convert to go from utf8 -> to our best guess at a useful
- * windows codepage.
+ * Returns an iconv converter to go from utf8 -> to our best guess at a useful
+ * 	windows codepage.
  **/
 GIConv
 gsf_msole_iconv_open_codepage_for_export (guint codepage_to)
@@ -1295,11 +1602,95 @@ gsf_msole_iconv_open_codepage_for_export (guint codepage_to)
 /**
  * gsf_msole_iconv_open_for_export :
  *
- * Open an iconv convert to go from utf8 -> to our best guess at a useful
- * windows codepage.
+ * Returns an iconv convert to go from utf8 -> to our best guess at a useful
+ * 	windows codepage.
  **/
 GIConv
 gsf_msole_iconv_open_for_export (void)
 {
 	return gsf_msole_iconv_open_codepage_for_export (gsf_msole_iconv_win_codepage ());
 }
+
+#define VBA_COMPRESSION_WINDOW 4096
+
+/**
+ * gsf_msole_inflate:
+ * @input: stream to read from
+ * @offset: offset into it for start byte of compresse stream
+ * 
+ * Decompresses an LZ compressed stream.
+ * 
+ * Return value: A GByteArray that the caller is responsible for freeing
+ **/
+GByteArray *
+gsf_msole_inflate (GsfInput *input, gsf_off_t offset)
+{
+	GByteArray *res;
+	unsigned	i, win_pos, pos = 0;
+	unsigned	mask, shift, distance;
+	guint8		flag, buffer [VBA_COMPRESSION_WINDOW];
+	guint8 const   *tmp;
+	guint16		token, len;
+	gboolean	clean = TRUE;
+
+	if (gsf_input_seek (input, offset, G_SEEK_SET))
+		return NULL;
+
+	res = g_byte_array_new ();
+
+	/* explaination from libole2/ms-ole-vba.c */
+	/* The first byte is a flag byte.  Each bit in this byte
+	 * determines what the next byte is.  If the bit is zero,
+	 * the next byte is a character.  Otherwise the  next two
+	 * bytes contain the number of characters to copy from the
+	 * umcompresed buffer and where to copy them from (offset,
+	 * length).
+	 */
+	while (NULL != gsf_input_read (input, 1, &flag))
+		for (mask = 1; mask < 0x100 ; mask <<= 1)
+			if (flag & mask) {
+				if (NULL == (tmp = gsf_input_read (input, 2, NULL)))
+					break;
+				win_pos = pos % VBA_COMPRESSION_WINDOW;
+				if (win_pos <= 0x80) {
+					if (win_pos <= 0x20)
+						shift = (win_pos <= 0x10) ? 12 : 11;
+					else
+						shift = (win_pos <= 0x40) ? 10 : 9;
+				} else {
+					if (win_pos <= 0x200)
+						shift = (win_pos <= 0x100) ? 8 : 7;
+					else if (win_pos <= 0x800)
+						shift = (win_pos <= 0x400) ? 6 : 5;
+					else
+						shift = 4;
+				}
+
+				token = GSF_LE_GET_GUINT16 (tmp);
+				len = (token & ((1 << shift) - 1)) + 3;
+				distance = token >> shift;
+				clean = TRUE;
+/*				fprintf (stderr, "Shift %d, token len %d, distance %d bytes %.2x %.2x\n",
+				shift, len, distance, (token & 0xff), (token >> 8)); */
+
+				for (i = 0; i < len; i++) {
+					unsigned srcpos = (pos - distance - 1) % VBA_COMPRESSION_WINDOW;
+					guint8 c = buffer [srcpos];
+					buffer [pos++ % VBA_COMPRESSION_WINDOW] = c;
+				}
+			} else {
+				if ((pos != 0) && ((pos % VBA_COMPRESSION_WINDOW) == 0) && clean) {
+					(void) gsf_input_read (input, 2, NULL);
+					clean = FALSE;
+					g_byte_array_append (res, buffer, VBA_COMPRESSION_WINDOW);
+					break;
+				}
+				if (NULL != gsf_input_read (input, 1, buffer + (pos % VBA_COMPRESSION_WINDOW)))
+					pos++;
+				clean = TRUE;
+			}
+
+	if (pos % VBA_COMPRESSION_WINDOW)
+		g_byte_array_append (res, buffer, pos % VBA_COMPRESSION_WINDOW);
+	return res;
+}
