@@ -231,12 +231,10 @@ gsf_output_gzip_write (GsfOutput *output,
 }
 
 static gboolean
-gsf_output_gzip_seek (GsfOutput *output, gsf_off_t offset, GSeekType whence)
+gsf_output_gzip_seek (G_GNUC_UNUSED GsfOutput *output,
+		      G_GNUC_UNUSED gsf_off_t offset,
+		      G_GNUC_UNUSED GSeekType whence)
 {
-	(void) output;
-	(void) offset;
-	(void) whence;
-	
 	return FALSE;
 }
 
