@@ -2,7 +2,7 @@
 /*
  * gsf-output-stdio.c: stdio based output
  *
- * Copyright (C) 2002-2003 Jody Goldberg (jody@gnome.org)
+ * Copyright (C) 2002-2004 Jody Goldberg (jody@gnome.org)
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of version 2.1 of the GNU Lesser General Public
@@ -28,12 +28,22 @@
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
+#ifdef HAVE_UNISTD_H
 #include <unistd.h>
+#endif
 #include <sys/types.h>
 #include <sys/stat.h>
 
+
 #ifdef G_OS_WIN32
 
+#ifndef S_IRUSR
+#define S_IRUSR 04
+#define S_IWUSR 02
+#define W_OK    00  /* existance mode only */
+typedef int mode_t;
+#endif
+
 #define S_IRGRP S_IRUSR
 #define S_IROTH S_IRUSR
 #define S_ISGID 0
@@ -42,14 +52,20 @@
 #define getgid() 0
 #define chown(filename, uid, gid) 0
 
+#ifdef HAVE_IO_H
+#include <io.h>
+#endif
+
 #endif /* G_OS_WIN32 */
 
+static GObjectClass *parent_class;
+
 struct _GsfOutputStdio {
 	GsfOutput output;
 
 	FILE     *file;
 	char	 *real_filename, *temp_filename;
-	gboolean  create_backup_copy;
+	gboolean  create_backup_copy, keep_open;
 	struct stat st;
 };
 
@@ -57,6 +73,21 @@ typedef struct {
 	GsfOutputClass output_class;
 } GsfOutputStdioClass;
 
+static int
+rename_wrapper (const char *oldfilename, const char *newfilename)
+{
+	int result = rename (oldfilename, newfilename);
+#ifdef G_OS_WIN32
+	if (result) {
+		/* Win32's rename does not unlink the target.  */
+		(void)unlink (newfilename);
+		result = rename (oldfilename, newfilename);
+	}
+#endif
+	return result;
+}
+
+
 #define GSF_MAX_LINK_LEVEL 256
 #ifdef MAXPATHLEN
 #define GSF_MAX_PATH_LEN  MAXPATHLEN
@@ -93,8 +124,9 @@ follow_symlinks (char const *filename, GError **error)
 
 			default:
 				if (error)
-					*error = g_error_new (gsf_output_error_id (), errno,
-							      g_strerror (errno));
+					*error = g_error_new_literal 
+						(gsf_output_error_id (), errno,
+						 g_strerror (errno));
 				g_free (followed_filename);
 				return NULL;
 			}
@@ -123,8 +155,8 @@ follow_symlinks (char const *filename, GError **error)
 
 	/* Too many symlinks */
 	if (error)
-		*error = g_error_new (gsf_output_error_id (), ELOOP,
-				      g_strerror (ELOOP));
+		*error = g_error_new_literal (gsf_output_error_id (), ELOOP,
+					      g_strerror (ELOOP));
 
 	return NULL;
 #else  /* !HAVE_READLINK */
@@ -141,19 +173,21 @@ follow_symlinks (char const *filename, GError **error)
  *
  * Returns a new file or NULL.
  **/
-GsfOutputStdio *
+GsfOutput *
 gsf_output_stdio_new (char const *filename, GError **err)
 {
 	GsfOutputStdio *stdio;
 	FILE *file = NULL;
-	char *dirname, *temp_filename = NULL;
+	char *dirname = NULL;
+	char *temp_filename = NULL;
 	char *real_filename = follow_symlinks (filename, err);
 	int fd;
 	mode_t saved_umask;
 	struct stat st;
-	
+	gboolean fixup_mode = FALSE;
+
 	if (real_filename == NULL)
-		return NULL;
+		goto failure;
 
 	/* Get the directory in which the real filename lives */
 	dirname = g_path_get_dirname (real_filename);
@@ -163,8 +197,9 @@ gsf_output_stdio_new (char const *filename, GError **err)
 		/* FIXME? Race conditions en masse.  */
 		if (access (real_filename, W_OK) != 0) {
 			if (err != NULL)
-				*err = g_error_new (gsf_output_error_id (), errno,
-						    g_strerror (errno));
+				*err = g_error_new_literal
+					(gsf_output_error_id (), errno,
+					 g_strerror (errno));
 			goto failure;
 		}
 	} else {
@@ -177,7 +212,7 @@ gsf_output_stdio_new (char const *filename, GError **err)
 		int result;
 
 		/* Use default permissions */
-		st.st_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
+		st.st_mode = 0666;  fixup_mode = TRUE;
 		st.st_uid = getuid ();
 
 		/* Used to set create_backup_copy = FALSE, but
@@ -196,16 +231,19 @@ gsf_output_stdio_new (char const *filename, GError **err)
 	 * implementations of mkstemp() use permissions 0666 and we want 0600.
 	 */
 	temp_filename = g_build_filename (dirname, ".gsf-save-XXXXXX", NULL);
-	g_free (dirname);
 	/* Oh, joy.  What about threads?  --MW */
 	saved_umask = umask (0077);
 	fd = g_mkstemp (temp_filename); /* this modifies temp_filename to the used name */
 	umask (saved_umask);
 
+	if (fixup_mode)
+		st.st_mode &= ~saved_umask;
+
 	if (fd < 0 || NULL == (file = fdopen (fd, "wb"))) {
 		if (err != NULL)
-			*err = g_error_new (gsf_output_error_id (), errno,
-					    g_strerror (errno));
+			*err = g_error_new_literal
+				(gsf_output_error_id (), errno,
+				 g_strerror (errno));
 		goto failure;
 	}
 
@@ -215,16 +253,50 @@ gsf_output_stdio_new (char const *filename, GError **err)
 	stdio->create_backup_copy = FALSE;
 	stdio->real_filename = real_filename;
 	stdio->temp_filename = temp_filename;
-	gsf_output_set_name (GSF_OUTPUT (stdio), filename);
 
-	return stdio;
+	gsf_output_set_name_from_filename (GSF_OUTPUT (stdio), filename);
+
+	g_free (dirname);
+
+	return GSF_OUTPUT (stdio);
 
 failure :
 	g_free (temp_filename);
 	g_free (real_filename);
+	g_free (dirname);
 	return NULL;
 }
 
+/**
+ * gsf_output_stdio_new_FILE :
+ * @filename  : The filename corresponding to @file.
+ * @file      : an existing stdio FILE *
+ * @keep_open : Should @file be closed when the wrapper is closed
+ *
+ * Assumes ownership of @file.  If @keep_open is true, ownership reverts
+ * to caller when the Gsfobject is closed.
+ *
+ * Returns a new GsfOutput wrapper for @file.  Warning: the result will be
+ * seekable only if @file is seekable.  If it is seekable, the resulting
+ * GsfOutput object will seek relative to @file's beginning, not its
+ * current location at the time the GsfOutput object is created.
+ **/
+GsfOutput *
+gsf_output_stdio_new_FILE (char const *filename, FILE *file, gboolean keep_open)
+{
+	GsfOutputStdio *stdio;
+
+	g_return_val_if_fail (filename != NULL, NULL);
+	g_return_val_if_fail (file != NULL, NULL);
+
+	stdio = g_object_new (GSF_OUTPUT_STDIO_TYPE, NULL);
+	stdio->file = file;
+	stdio->keep_open = keep_open;
+	stdio->real_filename = stdio->temp_filename = NULL;
+	gsf_output_set_name_from_filename (GSF_OUTPUT (stdio), filename);
+	return GSF_OUTPUT (stdio);
+}
+
 static gboolean
 gsf_output_stdio_close (GsfOutput *output)
 {
@@ -235,10 +307,29 @@ gsf_output_stdio_close (GsfOutput *output)
 	if (stdio->file == NULL)
 		return FALSE;
 
+	if (stdio->keep_open) {
+		gboolean res = (0 == fflush (stdio->file));
+		if (!res)
+			gsf_output_set_error (output, errno,
+					      "Failed to flush.");
+		stdio->file = NULL;
+		return res;
+	}
+
 	res = (0 == fclose (stdio->file));
 	stdio->file = NULL;
+
+	/* short circuit our when dealing with raw FILE */
+	if (NULL == stdio->real_filename) {
+		if (!res)
+			gsf_output_set_error (output, errno,
+				"Failed to close file.");
+		return res;
+	}
+
 	if (!res) {
-		gsf_output_set_error (output, errno, " ");
+		gsf_output_set_error (output, errno,
+				      "Failed to close temporary file.");
 		unlink (stdio->temp_filename);
 		return FALSE;
 	}
@@ -247,7 +338,7 @@ gsf_output_stdio_close (GsfOutput *output)
 	if (stdio->create_backup_copy) {
 		gint result;
 		backup_filename = g_strconcat (stdio->real_filename, ".bak", NULL);
-		result = rename (stdio->real_filename, backup_filename);
+		result = rename_wrapper (stdio->real_filename, backup_filename);
 		if (result != 0) {
 			char *utf8name = gsf_filename_to_utf8 (backup_filename, TRUE);
 			gsf_output_set_error (output, errno,
@@ -261,10 +352,10 @@ gsf_output_stdio_close (GsfOutput *output)
 	}
 
 	/* Move the temp file to the original file */
-	if (rename (stdio->temp_filename, stdio->real_filename) != 0) {
+	if (rename_wrapper (stdio->temp_filename, stdio->real_filename) != 0) {
 		gint saved_errno = errno;
 		if (backup_filename != NULL &&
-		    rename (backup_filename, stdio->real_filename) != 0)
+		    rename_wrapper (backup_filename, stdio->real_filename) != 0)
 			saved_errno = errno;
 		res = gsf_output_set_error (output, errno, g_strerror (errno));
 	} else {
@@ -291,7 +382,6 @@ gsf_output_stdio_close (GsfOutput *output)
 static void
 gsf_output_stdio_finalize (GObject *obj)
 {
-	GObjectClass *parent_class;
 	GsfOutput	*output = (GsfOutput *)obj;
 	GsfOutputStdio	*stdio = GSF_OUTPUT_STDIO (output);
 
@@ -303,9 +393,7 @@ gsf_output_stdio_finalize (GObject *obj)
 	g_free (stdio->temp_filename);
 	stdio->temp_filename = NULL;
 
-	parent_class = g_type_class_peek (GSF_OUTPUT_TYPE);
-	if (parent_class && parent_class->finalize)
-		parent_class->finalize (obj);
+	parent_class->finalize (obj);
 }
 
 static gboolean
@@ -313,28 +401,42 @@ gsf_output_stdio_seek (GsfOutput *output, gsf_off_t offset, GSeekType whence)
 {
 	GsfOutputStdio const *stdio = GSF_OUTPUT_STDIO (output);
 	int stdio_whence = 0;	/* make compiler shut up */
+
+#ifndef HAVE_FSEEKO
 	long loffset;
+#else
+	off_t loffset;
+#endif
 
 	g_return_val_if_fail (stdio->file != NULL, 
-		gsf_output_set_error (output, 0, "missing file"));
+			      gsf_output_set_error (output, 0, "missing file"));
 
 	loffset = offset;
 	if ((gsf_off_t) loffset != offset) { /* Check for overflow */
+#ifdef HAVE_FSEEKO
+		g_warning ("offset too large for fseeko");
+		return gsf_output_set_error (output, 0, "offset too large for fseeko");
+#else
 		g_warning ("offset too large for fseek");
 		return gsf_output_set_error (output, 0, "offset too large for fseek");
+#endif
 	}
 	switch (whence) {
+	default : ; /*checked in GsfOutput wrapper */
 	case G_SEEK_SET : stdio_whence = SEEK_SET;	break;
 	case G_SEEK_CUR : stdio_whence = SEEK_CUR;	break;
 	case G_SEEK_END : stdio_whence = SEEK_END;	break;
-	default :
-		break; /*checked in GsfOutput wrapper */
 	}
 
+	errno = 0;
+#ifdef HAVE_FSEEKO
+	if (0 == fseeko (stdio->file, loffset, stdio_whence))
+		return TRUE;
+#else
 	if (0 == fseek (stdio->file, loffset, stdio_whence))
 		return TRUE;
-	return gsf_output_set_error (output, errno,
-		g_strerror (errno));
+#endif
+	return gsf_output_set_error (output, errno, g_strerror (errno));
 }
 
 static gboolean
@@ -353,21 +455,18 @@ gsf_output_stdio_write (GsfOutput *output,
 	while (remaining > 0) {
 		written = fwrite (buffer + (num_bytes - remaining), 1, 
 				  remaining, stdio->file);
-		if ((written < remaining) && ferror (stdio->file) != 0) {
+		if ((written < remaining) && ferror (stdio->file) != 0)
 			return gsf_output_set_error (output, errno, g_strerror (errno));
-		}
+
 		remaining -= written;
 	}
 	return TRUE;
 }
 
-static gboolean
+static gsf_off_t
 gsf_output_stdio_vprintf (GsfOutput *output, char const *fmt, va_list args)
 {
-	GsfOutputStdio *stdio = (GsfOutputStdio *)output;
-	int res = vfprintf (stdio->file, fmt, args);
-
-	return (res >= 0);
+	return vfprintf (((GsfOutputStdio *)output)->file, fmt, args);
 }
 
 static void
@@ -377,6 +476,7 @@ gsf_output_stdio_init (GObject *obj)
 
 	stdio->file = NULL;
 	stdio->create_backup_copy = FALSE;
+	stdio->keep_open	  = FALSE;
 }
 
 static void
@@ -389,6 +489,8 @@ gsf_output_stdio_class_init (GObjectClass *gobject_class)
 	output_class->Seek	= gsf_output_stdio_seek;
 	output_class->Write	= gsf_output_stdio_write;
 	output_class->Vprintf	= gsf_output_stdio_vprintf;
+
+	parent_class = g_type_class_peek_parent (gobject_class);
 }
 
 GSF_CLASS (GsfOutputStdio, gsf_output_stdio,
