@@ -90,13 +90,8 @@ get_biff_opcode_name (unsigned opcode)
 	/* Count backwars to give preference to non-filtered record types */
 	for (lp=biff_types->len; --lp >= 0 ;) {
 		GENERIC_TYPE *bt = g_ptr_array_index (biff_types, lp);
-		if (bt->opcode>0xff) {
-			if (bt->opcode == opcode)
-				return bt->name;
-		} else {
-			if (bt->opcode == (opcode&0xff))
-				return bt->name;
-		}
+		if (bt->opcode == opcode)
+			return bt->name;
 	}
 	return "Unknown";
 }
@@ -125,9 +120,9 @@ dump_biff_stream (GsfInput *stream)
 		}
 
 		if (enable_dump)
-			printf ("Opcode 0x%3hx : %15s, length 0x%hx (=%hd) @ pos = 0x%x (=%d)\n",
+			printf ("Opcode 0x%3hx : %15s, length 0x%hx (=%hd)\n",
 				opcode, get_biff_opcode_name (opcode),
-				len, len, pos, pos);
+				len, len);
 
 		if (len > 0) {
 			data = gsf_input_read (stream, len, NULL);
@@ -139,6 +134,14 @@ dump_biff_stream (GsfInput *stream)
 		pos = gsf_input_tell (stream);
 	}
 }
+
+static void
+print_property (gpointer name, gpointer value, gpointer user_data)
+{
+	printf ("print_property: name  = %s\n", (char *)name);
+	printf ("                value = %s\n", g_strdup_value_contents((GValue *)value));
+}
+
 static int
 test (unsigned argc, char *argv[])
 {
@@ -151,11 +154,14 @@ test (unsigned argc, char *argv[])
 	GsfInfile *infile;
 	GError    *err = NULL;
 	unsigned i, j;
+	GsfDocMetaData *meta_data;
 
 	for (i = 1 ; i < argc ; i++) {
 		fprintf( stderr, "%s\n",argv[i]);
 
-		input = GSF_INPUT (gsf_input_mmap_new (argv[i], &err));
+		input = gsf_input_mmap_new (argv[i], NULL);
+		if (input == NULL)	/* Only report error if stdio fails too */
+			input = gsf_input_stdio_new (argv[i], &err);
 		if (input == NULL) {
 			g_return_val_if_fail (err != NULL, 1);
 			g_warning ("'%s' error: %s", argv[i], err->message);
@@ -165,7 +171,7 @@ test (unsigned argc, char *argv[])
 
 		input = gsf_input_uncompress (input);
 
-		infile = GSF_INFILE (gsf_infile_msole_new (input, &err));
+		infile = gsf_infile_msole_new (input, &err);
 
 		if (infile == NULL) {
 
@@ -194,24 +200,36 @@ test (unsigned argc, char *argv[])
 		stream = gsf_infile_child_by_name (infile, "\05SummaryInformation");
 		if (stream != NULL) {
 			puts ( "SummaryInfo");
-			gsf_msole_metadata_read (stream, &err);
+			meta_data = gsf_msole_metadata_read_real (stream, &err);
 			if (err != NULL)
 				g_warning ("'%s' error: %s", argv[i], err->message);
+			else {
+				gsf_doc_meta_data_foreach (meta_data, print_property, NULL);
+				g_object_unref (meta_data);
+			}
+
 			g_object_unref (G_OBJECT (stream));
+			
 		}
 
 		stream = gsf_infile_child_by_name (infile, "\05DocumentSummaryInformation");
 		if (stream != NULL) {
 			puts ( "DocSummaryInfo");
-			gsf_msole_metadata_read (stream, &err);
+			meta_data = gsf_msole_metadata_read_real (stream, &err);
 			if (err != NULL)
 				g_warning ("'%s' error: %s", argv[i], err->message);
+			else {
+				gsf_doc_meta_data_foreach (meta_data, print_property, NULL);
+				g_object_unref (meta_data);
+			}
+			
 			g_object_unref (G_OBJECT (stream));
 		}
 
 		for (j = 0 ; j < G_N_ELEMENTS (stream_names) ; j++) {
 			stream = gsf_infile_child_by_name (infile, stream_names[j]);
 			if (stream != NULL) {
+				puts (j < 3 ? "Excel97" : "Excel95");
 				dump_biff_stream (stream);
 				g_object_unref (G_OBJECT (stream));
 				break;
